# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .convert import get_message_slot_types
from .convert import message_to_csv
from .convert import message_to_ordereddict
from .convert import message_to_yaml
from .get_interfaces import get_action_interfaces
from .get_interfaces import get_interface_packages
from .get_interfaces import get_interface_path
from .get_interfaces import get_interfaces
from .get_interfaces import get_message_interfaces
from .get_interfaces import get_service_interfaces
from .import_message import import_message_from_namespaced_type
from .set_message import set_message_fields


__all__ = [
    'get_action_interfaces',
    'get_interface_packages',
    'get_message_interfaces',
    'get_service_interfaces',
    'get_interface_path',
    'get_interfaces',
    'get_message_slot_types',
    'import_message_from_namespaced_type',
    'message_to_csv',
    'message_to_ordereddict',
    'message_to_yaml',
    'set_message_fields',
]



# Copyright 2016-2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
from collections import OrderedDict
import sys
from typing import Any

import numpy
import rosidl_parser.definition
import yaml


__yaml_representer_registered = False


def __get_type_name(value_type):
    if isinstance(value_type, rosidl_parser.definition.BasicType):
        return value_type.typename
    elif isinstance(value_type, rosidl_parser.definition.AbstractString):
        return 'string'
    elif isinstance(value_type, rosidl_parser.definition.AbstractWString):
        return 'wstring'
    elif isinstance(value_type, rosidl_parser.definition.NamedType):
        return value_type.name
    elif isinstance(value_type, rosidl_parser.definition.NamespacedType):
        return '/'.join(value_type.namespaced_name())
    else:
        return 'unknown'


def __abbreviate_array_info(value, field_type):
    value_type_name = __get_type_name(field_type.value_type)
    if isinstance(field_type, rosidl_parser.definition.Array):
        return '<array type: {0}[{1}]>'.format(
            value_type_name, field_type.size)
    elif isinstance(field_type, rosidl_parser.definition.BoundedSequence):
        return '<sequence type: {0}[{1}], length: {2}>'.format(
            value_type_name, field_type.maximum_size, len(value))
    elif isinstance(field_type, rosidl_parser.definition.UnboundedSequence):
        return '<sequence type: {0}, length: {1}>'.format(
            value_type_name, len(value))
    return 'unknown'


# Custom representer for getting clean YAML output that preserves the order in an OrderedDict.
# Inspired by: http://stackoverflow.com/a/16782282/7169408
def __represent_ordereddict(dumper, data):
    items = []
    for k, v in data.items():
        items.append((dumper.represent_data(k), dumper.represent_data(v)))
    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', items)


def message_to_yaml(
    msg: Any,
    *,
    truncate_length: int = None,
    no_arr: bool = False,
    no_str: bool = False,
    flow_style: bool = False
) -> str:
    """
    Convert a ROS message to a YAML string.

    :param msg: The ROS message to convert.
    :param truncate_length: Truncate values for all message fields to this length.
        This does not truncate the list of message fields.
    :param no_arr: Exclude array fields of the message.
    :param no_str: Exclude string fields of the message.
    :returns: A YAML string representation of the input ROS message.
    """
    global __yaml_representer_registered

    # Register our custom representer for YAML output
    if not __yaml_representer_registered:
        yaml.add_representer(OrderedDict, __represent_ordereddict)
        __yaml_representer_registered = True

    return yaml.dump(
        message_to_ordereddict(
            msg, truncate_length=truncate_length, no_arr=no_arr, no_str=no_str),
        allow_unicode=True, width=sys.maxsize, default_flow_style=flow_style,
    )


def message_to_csv(
    msg: Any,
    *,
    truncate_length: int = None,
    no_arr: bool = False,
    no_str: bool = False
) -> str:
    """
    Convert a ROS message to string of comma-separated values.

    :param msg: The ROS message to convert.
    :param truncate_length: Truncate values for all message fields to this length.
        This does not truncate the list of message fields.
    :param no_arr: Exclude array fields of the message.
    :param no_str: Exclude string fields of the message.
    :returns: A string of comma-separated values representing the input message.
    """
    def to_string(val, field_type=None):
        nonlocal truncate_length, no_arr, no_str
        r = ''
        if any(isinstance(val, t) for t in [list, tuple, array.array, numpy.ndarray]):
            if no_arr is True and field_type is not None:
                r = __abbreviate_array_info(val, field_type)
            else:
                for i, v in enumerate(val):
                    if r:
                        r += ','
                    if truncate_length is not None and i >= truncate_length:
                        r += '...'
                        break
                    r += to_string(v)
        elif any(isinstance(val, t) for t in [bool, bytes, float, int, str, numpy.number]):
            if no_str is True and isinstance(val, str):
                val = '<string length: <{0}>>'.format(len(val))
            elif any(isinstance(val, t) for t in [bytes, str]):
                if truncate_length is not None and len(val) > truncate_length:
                    val = val[:truncate_length]
                    if isinstance(val, bytes):
                        val += b'...'
                    else:
                        val += '...'
            r = str(val)
        else:
            r = message_to_csv(val, truncate_length=truncate_length, no_arr=no_arr, no_str=no_str)
        return r
    result = ''

    # We rely on __slots__ retaining the order of the fields in the .msg file.
    for field_name, field_type in zip(msg.__slots__, msg.SLOT_TYPES):
        value = getattr(msg, field_name)

        if result:
            result += ','

        result += to_string(value, field_type)
    return result


# Convert a msg to an OrderedDict. We do this instead of implementing a generic __dict__() method
# in the msg because we want to preserve order of fields from the .msg file(s).
def message_to_ordereddict(
    msg: Any,
    *,
    truncate_length: int = None,
    no_arr: bool = False,
    no_str: bool = False
) -> OrderedDict:
    """
    Convert a ROS message to an OrderedDict.

    :param msg: The ROS message to convert.
    :param truncate_length: Truncate values for all message fields to this length.
        This does not truncate the list of fields (ie. the dictionary keys).
    :param no_arr: Exclude array fields of the message.
    :param no_str: Exclude string fields of the message.
    :returns: An OrderedDict where the keys are the ROS message fields and the values are
        set to the values of the input message.
    """
    d = OrderedDict()

    # We rely on __slots__ retaining the order of the fields in the .msg file.
    for field_name, field_type in zip(msg.__slots__, msg.SLOT_TYPES):
        value = getattr(msg, field_name, None)

        value = _convert_value(
            value, field_type=field_type,
            truncate_length=truncate_length, no_arr=no_arr, no_str=no_str)
        # Remove leading underscore from field name
        d[field_name[1:]] = value
    return d


def _convert_value(
    value,
    *,
    field_type=None,
    truncate_length=None,
    no_arr=False,
    no_str=False
):

    if isinstance(value, bytes):
        if truncate_length is not None and len(value) > truncate_length:
            value = ''.join([chr(c) for c in value[:truncate_length]]) + '...'
        else:
            value = ''.join([chr(c) for c in value])
    elif isinstance(value, str):
        if no_str is True:
            value = '<string length: <{0}>>'.format(len(value))
        elif truncate_length is not None and len(value) > truncate_length:
            value = value[:truncate_length] + '...'
    elif isinstance(value, (list, tuple, array.array, numpy.ndarray)):
        # Since arrays and ndarrays can't contain mixed types convert to list
        typename = tuple if isinstance(value, tuple) else list
        if no_arr is True and field_type is not None:
            value = __abbreviate_array_info(value, field_type)
        elif truncate_length is not None and len(value) > truncate_length:
            # Truncate the sequence
            value = value[:truncate_length]
            # Truncate every item in the sequence
            value = typename(
                [_convert_value(v, truncate_length=truncate_length,
                                no_arr=no_arr, no_str=no_str) for v in value] + ['...'])
        else:
            # Truncate every item in the list
            value = typename(
                [_convert_value(v, truncate_length=truncate_length,
                                no_arr=no_arr, no_str=no_str) for v in value])
    elif isinstance(value, dict) or isinstance(value, OrderedDict):
        # Convert each key and value in the mapping
        new_value = {} if isinstance(value, dict) else OrderedDict()
        for k, v in value.items():
            # Don't truncate keys because that could result in key collisions and data loss
            new_value[_convert_value(k)] = _convert_value(
                v, truncate_length=truncate_length, no_arr=no_arr, no_str=no_str)
        value = new_value
    elif isinstance(value, numpy.number):
        value = value.item()
    elif not isinstance(value, (bool, float, int)):
        # Assuming value is a message since it is neither a collection nor a primitive type
        value = message_to_ordereddict(
            value, truncate_length=truncate_length, no_arr=no_arr, no_str=no_str)
    return value


def get_message_slot_types(msg: Any) -> OrderedDict:
    """
    Return an OrderedDict of the slot types of a message.

    :param msg: The ROS message to get members types from.
    :returns: An OrderedDict with message member names as keys and slot types as values.
    """
    return OrderedDict(zip([s[1:] for s in msg.__slots__], msg.SLOT_TYPES))



# Copyright 2019 Mikael Arguedas.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from typing import Any
import warnings

from rosidl_parser.definition import NamespacedType


def import_message_from_namespaced_type(message_type: NamespacedType) -> Any:
    if not isinstance(message_type, NamespacedType):
        if hasattr(message_type, 'value_type'):
            message_type = message_type.value_type
            warnings.warn(
                'Passing objects containing a NamespacedType is deprecated, '
                'please pass a NamespacedType object directly',
                DeprecationWarning)
    module = importlib.import_module(
        '.'.join(message_type.namespaces))

    # Special case for action feedback
    if message_type.name.endswith('_FeedbackMessage'):
        action_name, name = message_type.name.rsplit('_', 1)
        return getattr(getattr(getattr(module, action_name), 'Impl'), name)

    return getattr(module, message_type.name)



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Text

from rosidl_parser.definition import NamespacedType
from rosidl_runtime_py.import_message import import_message_from_namespaced_type


def get_interface(identifier: Text):
    """Get an interface from its full name."""
    return import_message_from_namespaced_type(get_namespaced_type(identifier))


def get_message(identifier: Text):
    """Get a message from its full name."""
    interface = import_message_from_namespaced_type(get_message_namespaced_type(identifier))
    if not is_message(interface):
        raise ValueError("Expected the full name of a message, got '{}'".format(identifier))
    return interface


def get_service(identifier: Text):
    """Get a service from its full name."""
    interface = import_message_from_namespaced_type(get_service_namespaced_type(identifier))
    if not is_service(interface):
        raise ValueError("Expected the full name of a service, got '{}'".format(identifier))
    return interface


def get_action(identifier: Text):
    """Get a message from its full name."""
    interface = import_message_from_namespaced_type(get_action_namespaced_type(identifier))
    if not is_action(interface):
        raise ValueError("Expected the full name of an action, got '{}'".format(identifier))
    return interface


def get_namespaced_type(identifier: Text):
    """Create a `NamespacedType` object from the full name of an interface."""
    return _get_namespaced_type(identifier)


def get_message_namespaced_type(identifier: Text):
    """Create a `NamespacedType` object from the full name of a message."""
    return _get_namespaced_type(identifier, default_namespace='msg')


def get_service_namespaced_type(identifier: Text):
    """Create a `NamespacedType` object from the full name of a service."""
    return _get_namespaced_type(identifier, default_namespace='srv')


def get_action_namespaced_type(identifier: Text):
    """Create a `NamespacedType` object from the full name of an action."""
    return _get_namespaced_type(identifier, default_namespace='action')


def is_message(interface):
    """Return `True` if `interface` is a message."""
    return hasattr(interface, 'SLOT_TYPES')


def is_service(interface):
    """Return `True` if `interface` is a service."""
    return hasattr(interface, 'Request') and hasattr(interface, 'Response')


def is_action(interface):
    """Return `True` if `interface` is an action."""
    return (
        hasattr(interface, 'Goal') and
        hasattr(interface, 'Result') and
        hasattr(interface, 'Feedback')
    )


def _get_namespaced_type(identifier: Text, *, default_namespace=None):
    namespace = identifier.split(sep='/')
    name = namespace[-1]
    namespace = namespace[0:-1]
    if len(namespace) == 1 and default_namespace is not None:
        namespace.append(default_namespace)
    return NamespacedType(namespace, name)



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Dict
from typing import Iterable
from typing import List

from ament_index_python import get_resource
from ament_index_python import get_resources
from ament_index_python import has_resource


def get_interface_packages() -> List[str]:
    """Get all packages that generate interfaces."""
    return get_resources('rosidl_interfaces')


def _get_interfaces(package_names: Iterable[str] = []) -> Dict[str, List[str]]:
    interfaces = {}
    if len(package_names) == 0:
        package_names = get_resources('rosidl_interfaces')

    for package_name in package_names:
        if not has_resource('packages', package_name):
            raise LookupError(f"Unknown package '{package_name}'")
        try:
            content, _ = get_resource('rosidl_interfaces', package_name)
        except LookupError:
            continue
        interfaces[package_name] = content.splitlines()
    return interfaces


def get_interfaces(package_names: Iterable[str] = []) -> Dict[str, List[str]]:
    """
    Get interfaces for one or more packages.

    If a package does not contain any interfaces, then it will not exist in the returned
    dictionary.

    :param package_names: Interfaces are returned for these packages.
      If no package names are provided, then this function returns interfaces for all packages.
    :return: A dictionary where keys are package names and values are lists of interface names.
    :raises LookupError: If one or more packages can not be found.
    """
    interfaces = _get_interfaces(package_names)
    # filter out hidden interfaces
    filtered_interfaces = {}
    for package_name, interface_names in interfaces.items():
        _interfaces = list({
            interface_name.rsplit('.', 1)[0]
            for interface_name in interface_names
            if '_' not in interface_name
        })
        if _interfaces:
            filtered_interfaces[package_name] = _interfaces
    return filtered_interfaces


def get_message_interfaces(package_names: Iterable[str] = []) -> Dict[str, List[str]]:
    """
    Get message interfaces for one or more packages.

    If a package does not contain any message interfaces, then it will not exist in the returned
    dictionary.

    :param package_names: Message interfaces are returned for these packages.
      If no package names are provided, then this function returns interfaces for all packages.
    :return: A dictionary where keys are package names and values are lists of interface names.
    :raises LookupError: If one or more packages can not be found.
    """
    interfaces = _get_interfaces(package_names)
    # filter out hidden interfaces and identify message interfaces by the namespace and suffix
    filtered_interfaces = {}
    for package_name, interface_names in interfaces.items():
        message_interfaces = list({
            interface_name[:-4]
            for interface_name in interface_names
            if '_' not in interface_name and
            interface_name.startswith('msg/') and
            interface_name[-4:] in ('.idl', '.msg')
        })
        if message_interfaces:
            filtered_interfaces[package_name] = message_interfaces
    return filtered_interfaces


def get_service_interfaces(package_names: Iterable[str] = []) -> Dict[str, List[str]]:
    """
    Get service interfaces for one or more packages.

    If a package does not contain any service interfaces, then it will not exist in the returned
    dictionary.

    :param package_names: Service interfaces are returned for these packages.
      If no package names are provided, then this function returns interfaces for all packages.
    :return: A dictionary where keys are package names and values are lists of interface names.
    :raises LookupError: If one or more packages can not be found.
    """
    interfaces = _get_interfaces(package_names)
    # filter out hidden interfaces and identify service interfaces by the namespace and suffix
    filtered_interfaces = {}
    for package_name, interface_names in interfaces.items():
        service_interfaces = list({
            interface_name[:-4]
            for interface_name in interface_names
            if '_' not in interface_name and
            interface_name.startswith('srv/') and
            interface_name[-4:] in ('.idl', '.srv')
        })
        if service_interfaces:
            filtered_interfaces[package_name] = service_interfaces
    return filtered_interfaces


def get_action_interfaces(package_names: Iterable[str] = []) -> Dict[str, List[str]]:
    """
    Get action interfaces for one or more packages.

    If a package does not contain any action interfaces, then it will not exist in the returned
    dictionary.

    :param package_names: Action interfaces are returned for these packages.
      If no package names are provided, then this function returns interfaces for all packages.
    :return: A dictionary where keys are package names and values are lists of interface names.
    :raises LookupError: If one or more packages can not be found.
    """
    interfaces = _get_interfaces(package_names)
    # filter out hidden interfaces and identify action interfaces by the namespace and suffix
    filtered_interfaces = {}
    for package_name, interface_names in interfaces.items():
        action_interfaces = list({
            interface_name.rsplit('.', 1)[0]
            for interface_name in interface_names
            if '_' not in interface_name and
            interface_name.startswith('action/') and
            (interface_name[-4:] == '.idl' or interface_name[-7:] == '.action')
        })
        if action_interfaces:
            filtered_interfaces[package_name] = action_interfaces
    return filtered_interfaces


def get_interface_path(interface_name: str) -> str:
    """
    Get the path to an interface definition file.

    :param interface_name: The name of the interface (e.g. builtin_interfaces/msg/Time.msg)
      If no dot-separated suffix is provided, then it is inferred from the namespace.
    :return: The path to the interface definition file.
    :raises ValueError: If the interface name is malformed.
    :raises LookupError: If the package or interface cannot be found.
    """
    # Split up the name for analysis
    parts = interface_name.split('/')
    # By convention we expect there to be at least two parts to the interface
    if len(parts) < 2:
        raise ValueError(
            f"Invalid name '{interface_name}'. Expected at least two parts separated by '/'")
    if not all(parts):
        raise ValueError(f"Invalid name '{interface_name}'. Must not contain empty parts")
    if '..' in parts:
        raise ValueError(f"Invalid name '{interface_name}'. Must not contain '..'")
    # By convention we expect the first part to be the package name
    prefix_path = has_resource('packages', parts[0])
    if not prefix_path:
        raise LookupError(f"Unknown package '{parts[0]}'")

    interface_path = os.path.join(prefix_path, 'share', interface_name)
    # Check if there is a dot-separated suffix
    if len(parts[-1].rsplit('.', 1)) == 1:
        # If there is no suffix, try appending parent namespace (e.g. '.msg', '.srv', '.action')
        interface_path_with_suffix = interface_path + '.' + parts[-2]
        if os.path.exists(interface_path_with_suffix):
            return interface_path_with_suffix
        # Finally, try appending '.idl'
        interface_path += '.idl'

    if not os.path.exists(interface_path):
        raise LookupError(f"Could not find the interface '{interface_path}'")
    return interface_path



# Copyright 2017-2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array

from typing import Any
from typing import Dict

import numpy

from rosidl_parser.definition import AbstractNestedType
from rosidl_parser.definition import NamespacedType
from rosidl_runtime_py.convert import get_message_slot_types
from rosidl_runtime_py.import_message import import_message_from_namespaced_type


def set_message_fields(msg: Any, values: Dict[str, str]) -> None:
    """
    Set the fields of a ROS message.

    :param msg: The ROS message to populate.
    :param values: The values to set in the ROS message. The keys of the dictionary represent
        fields of the message.
    :raises AttributeError: If the message does not have a field provided in the input dictionary.
    :raises TypeError: If a message value does not match its field type.
    """
    try:
        items = values.items()
    except AttributeError:
        raise TypeError(
            "Value '%s' is expected to be a dictionary but is a %s" %
            (values, type(values).__name__))
    for field_name, field_value in items:
        field = getattr(msg, field_name)
        field_type = type(field)
        if field_type is array.array:
            value = field_type(field.typecode, field_value)
        elif field_type is numpy.ndarray:
            value = numpy.array(field_value, dtype=field.dtype)
        elif type(field_value) is field_type:
            value = field_value
        else:
            try:
                value = field_type(field_value)
            except TypeError:
                value = field_type()
                set_message_fields(value, field_value)
        rosidl_type = get_message_slot_types(msg)[field_name]
        # Check if field is an array of ROS messages
        if isinstance(rosidl_type, AbstractNestedType):
            if isinstance(rosidl_type.value_type, NamespacedType):
                field_elem_type = import_message_from_namespaced_type(rosidl_type.value_type)
                for n in range(len(value)):
                    submsg = field_elem_type()
                    set_message_fields(submsg, value[n])
                    value[n] = submsg
        setattr(msg, field_name, value)
